// Application : DPMITEST
// Description : DOS Protected Mode Interface function library test application
// Module      : DPMITEST.CPP
// Version     : 1.00 - 13/09/1994
// Written by  : David A. Mair
// -----------------------------------------------------------------------------
// Change summary --------------------------------------------------------------
// -----------------------------------------------------------------------------
//    Date   | Version | Comment and author
// -----------------------------------------------------------------------------
//           |         |
//           |         |
// -----------------------------------------------------------------------------
#pragma inline


#include <stdio.h>
#include "dpmi.h"


void Copyright(void)
{
	printf("DPMITEST v1.0 - Test application for DPMI library\n");
}


BOOL TestCPUMode(void)
{
	// What is the CPU operating mode
	if (InRealMode)
	{
		printf("CPU is operating in real or Virtual 8086 mode\n");

		return FALSE;
	}
	else
	{
		printf("CPU is operating in protected mode\n");
	}

	return TRUE;
}


void ShowDescriptorDetail(SELECTOR Selector)
{
	DWORD dwBaseAddress;
	DWORD dwLimit;
	ACCESS_RIGHTS byRights;
	EXTENDED_ACCESS_RIGHTS by386Rights;

	if (GetSegmentBaseAddress(Selector, &dwBaseAddress))
	{
		printf("Failed to load segment base address\n");
	}

	dwLimit = GetSegmentLimit(Selector);

	// Get the access rights for our code segment
	byRights = GetDescriptorAccessRights(Selector, &by386Rights);
			
	printf("Segment %X attributes\n", Selector);
	printf("       Table : %s\n", (Selector & LDT_SELECTOR) ? "LDT" : "GDT");
	printf("         RPL : %X\n", (Selector & SELECTOR_RPL));
	printf("Base Address : %X\n", dwBaseAddress);
	printf("       Limit : %X\n", dwLimit);
	printf("     Present : %s\n", byRights.Present ? "YES" : "NO");
	printf("         DPL : %X\n", byRights.DPL);
	printf("     Content : %s\n", byRights.CodeSegment ? "CODE" : "DATA");
	printf("      Expand : %s\n", byRights.ExpandDown ? "DOWN" : "UP");
	printf("      Access : %s\n", byRights.ReadWrite ? "READ/WRITE" : "READ ONLY");
	printf("        Used : %s\n", byRights.Accessed ? "YES" : "NO");
	printf("   Available : %s\n", by386Rights.AVL ? "YES" : "NO");
	printf("   Word Size : %d bit\n", by386Rights.Default32Bit ? 32 : 16);
	printf(" Granularity : %s\n", by386Rights.PageGranular ? "PAGE" : "BYTE");
}


void TestLDTSelectors(void)
{
	SELECTOR Selector;

	// Allocate a selector
	if (!AllocateLDTDescriptors(1, &Selector))
	{
		printf("Selector %4X allocated\n", Selector);

		if (!SetSegmentBaseAddress(Selector, 0x00000400))
		{
			printf("Segment base address set to 0x00000400\n");
		}
		else
		{
			printf("Failed to set segment base address\n");
		}

		if (!SetSegmentLimit(Selector, 0x000010000))
		{
			printf("Segment Limit set to 0x000010000\n");
		}
		else
		{
			printf("Failed to set segment limit\n");
		}

		ShowDescriptorDetail(Selector);

		// Free the descriptor
		if (!FreeLDTDescriptor(Selector))
		{
			printf("Selector %4X freed in the LDT\n", Selector);
		}
		else
		{
			printf("Failed to free selector %4huX in the LDT\n", Selector);
		}
	}
	else
	{
		printf("Failed to allocate a LDT descriptor\n");
	}
}


void TestCodeSegment(void)
{
	SELECTOR CodeSegment;

	// get the selector for our code segment
	asm {
		mov		ax, cs
		mov		word ptr [CodeSegment], ax
	}

	printf("Code segment selector is %X\n", CodeSegment);

	ShowDescriptorDetail(CodeSegment);
}


void TestAliasDescriptor(void)
{
	SELECTOR CodeSegment;
	SELECTOR AliasSelector;

	// get the selector for our code segment
	asm {
		mov		ax, cs
		mov		word ptr [CodeSegment], ax
	}

	if (!CreateAliasDescriptor(CodeSegment, &AliasSelector))
	{
		printf("Alias descriptor created, selector is %Xh\n", AliasSelector);
		
		ShowDescriptorDetail(AliasSelector);

		if (!FreeLDTDescriptor(AliasSelector))
		{
			printf("Alias descriptor freed\n");
		}
		else
		{
			printf("Failed to free alias descriptor\n");
		}
	}
	else
	{
		printf("Failed to create alias descriptor\n");
	}
}


void TestRealModeDescriptor(void)
{
	SELECTOR Selector;
	
	// Now create a real mode segment descriptor
	if (!SegmentToDescriptor(0xA000, &Selector))
	{
		printf("Selector %X allocated for real mode segment A000h\n", Selector);

		ShowDescriptorDetail(Selector);
	}
	else
	{
		printf("Failed to allocate a descriptor for real mode segment A000h\n");
	}
}


void TestSpecificLDTDescriptor(void)
{
	SELECTOR Selector;

	Selector = 0x007C;

	if (!AllocateSpecificLDTDescriptor(Selector))
	{
		printf("Selector %Xh specifically allocated\n", Selector);
		ShowDescriptorDetail(Selector);

		if (!FreeLDTDescriptor(Selector))
		{
			printf("Freed selector %Xh\n", Selector);
		}
		else
		{
			printf("Failed to free selector %Xh\n", Selector);
		}
	}
	else
	{
		printf("Failed to allocate Selector %Xh\n", Selector);
	}
}


void ShowDescriptorContent(SELECTOR Selector, DESCRIPTOR Descriptor)
{
	printf("Descriptor content for selector %Xh\n", Selector);

	// Find out descriptor type
	if (Descriptor.SegmentType.SystemSegment.EqualsZero == 0)
	{
		printf("Descriptor defines a System Segment\n");
		printf("       Limit  (0 - 15) : %Xh\n", Descriptor.Limit0to15);
		printf("       Limit (16 - 20) : %Xh\n", Descriptor.Limit16to20);
		printf("Base Address  (0 - 15) : %Xh\n", Descriptor.BaseAddress0to15);
		printf("Base Address (16 - 23) : %Xh\n", Descriptor.BaseAddress16to23);
		printf("Base Address (24 - 31) : %Xh\n", Descriptor.BaseAddress24to31);
		printf("                  Type : ");
		switch (Descriptor.SegmentType.SystemSegment.Type)
		{
			case 0:
				printf("RESERVED\n");
				break;
			case 1:
				printf("AVAILABLE 286 TSS\n");
				break;
			case 2:
				printf("LDT\n");
				break;
			case 3:
				printf("BUSY 286 TSS\n");
				break;
			case 4:
				printf("CALL GATE\n");
				break;
			case 5:
				printf("TASK GATE\n");
				break;
			case 6:
				printf("286 INTERRUPT GATE\n");
				break;
			case 7:
				printf("286 TRAP GATE\n");
				break;
			case 8:
				printf("RESERVED\n");
				break;
			case 9:
				printf("AVAILABLE i486 CPU TSS\n");
				break;
			case 10:
				printf("RESERVED\n");
				break;
			case 11:
				printf("BUSY i486 CPU TSS\n");
				break;
			case 12:
				printf("i486 CPU CALL GATE\n");
				break;
			case 13:
				printf("RESERVED\n");
				break;
			case 14:
				printf("i486 CPU INTERRUPT GATE\n");
				break;
			case 15:
				printf("i486 CPU TASK GATE\n");
				break;
		}
		printf("                   DPL : %Xh\n", Descriptor.SegmentType.SystemSegment.DPL);
		printf("               Present : %s\n", Descriptor.SegmentType.SystemSegment.Present ? "YES" : "NO");
	}
	else if (Descriptor.SegmentType.CodeSegment.EqualsThree == 3)
	{
		printf("Descriptor defines a Code Segment\n");
		printf("       Limit  (0 - 15) : %Xh\n", Descriptor.Limit0to15);
		printf("       Limit (16 - 20) : %Xh\n", Descriptor.Limit16to20);
		printf("Base Address  (0 - 15) : %Xh\n", Descriptor.BaseAddress0to15);
		printf("Base Address (16 - 23) : %Xh\n", Descriptor.BaseAddress16to23);
		printf("Base Address (24 - 31) : %Xh\n", Descriptor.BaseAddress24to31);
		printf("              Accessed : %s\n", Descriptor.SegmentType.CodeSegment.Accessed ? "YES" : "NO");
		printf("              Readable : %s\n", Descriptor.SegmentType.CodeSegment.Readable ? "YES" : "NO");
		printf("            Conforming : %s\n", Descriptor.SegmentType.CodeSegment.Conforming ? "YES" : "NO");
		printf("                   DPL : %Xh\n", Descriptor.SegmentType.CodeSegment.DPL);
		printf("               Present : %s\n", Descriptor.SegmentType.CodeSegment.Present ? "YES" : "NO");
	}
	else if (Descriptor.SegmentType.DataSegment.EqualsTwo == 2)
	{
		printf("Descriptor desfines a Data Segment\n");
		printf("       Limit  (0 - 15) : %Xh\n", Descriptor.Limit0to15);
		printf("       Limit (16 - 20) : %Xh\n", Descriptor.Limit16to20);
		printf("Base Address  (0 - 15) : %Xh\n", Descriptor.BaseAddress0to15);
		printf("Base Address (16 - 23) : %Xh\n", Descriptor.BaseAddress16to23);
		printf("Base Address (24 - 31) : %Xh\n", Descriptor.BaseAddress24to31);
		printf("              Accessed : %s\n", Descriptor.SegmentType.DataSegment.Accessed ? "YES" : "NO");
		printf("              Writable : %s\n", Descriptor.SegmentType.DataSegment.Writable ? "YES" : "NO");
		printf("           Expand-Down : %s\n", Descriptor.SegmentType.DataSegment.ExpandDown ? "YES" : "NO");
		printf("                   DPL : %Xh\n", Descriptor.SegmentType.DataSegment.DPL);
		printf("               Present : %s\n", Descriptor.SegmentType.DataSegment.Present ? "YES" : "NO");
	}
	else
	{
		printf("Descriptor defines an unknown segment type\n");
	}
	
	printf("             Available : %s\n", Descriptor.Available ? "YES" : "NO");
	printf("          Operand Size : %s bit\n", Descriptor.OperandSize ? "32" : "16");
	printf("           Granularity : %s\n", Descriptor.Granularity ? "PAGE" : "BYTE");
}


void TestGetAndSetDescriptor(void)
{
	SELECTOR CodeSegment;
	SELECTOR AliasSelector;
	DESCRIPTOR Descriptor;

	// get the selector for our code segment
	asm {
		mov		ax, cs
		mov		word ptr [CodeSegment], ax
	}

	if (!CreateAliasDescriptor(CodeSegment, &AliasSelector))
	{
		printf("Alias descriptor created, selector is %Xh\n", AliasSelector);
		
		ShowDescriptorDetail(AliasSelector);

		if (!GetDescriptor(AliasSelector, &Descriptor))
		{
			ShowDescriptorContent(AliasSelector, Descriptor);

			// Make a change to the descriptor
			Descriptor.SegmentType.DataSegment.ExpandDown = !Descriptor.SegmentType.DataSegment.ExpandDown;

			if (!SetDescriptor(AliasSelector, Descriptor))
			{
				if (!GetDescriptor(AliasSelector, &Descriptor))
				{
					ShowDescriptorContent(AliasSelector, Descriptor);
				}
				else
				{
					printf("Failed to get descriptor back\n");
				}
			}
			else
			{
				printf("Failed to set descriptor\n");
			}
		}
		else
		{
			printf("Failed to get descriptor\n");
		}

		if (!FreeLDTDescriptor(AliasSelector))
		{
			printf("Alias descriptor freed\n");
		}
		else
		{
			printf("Failed to free alias descriptor\n");
		}
	}
	else
	{
		printf("Failed to create alias descriptor\n");
	}
}


void TestMultipleGetAndSet(void)
{
	SELECTOR ASegment;
	SELECTOR ASelector;
	DESCRIPTOR_LIST Descriptors[2];
	WORD DescriptorCount;

	// get the selector for our code segment
	asm {
		mov		ax, cs
		mov		word ptr [ASelector], ax
	}

	// Alias it
	if (CreateAliasDescriptor(ASelector, &(Descriptors[0].Selector)))
	{
		printf("Failed to create alias descriptor\n");
		return;
	}
	
	// Get the selector for our data segment
	asm {
		mov		ax, ds
		mov		word ptr [ASelector], ax
	}

	// Alias it
	if (CreateAliasDescriptor(ASelector, &(Descriptors[1].Selector)))
	{
		printf("Failed to create alias descriptor\n");
		return;
	}
	
	DescriptorCount = 2;

	if (!GetMultipleDescriptors(&DescriptorCount, &(Descriptors[0].Selector)))
	{
		printf("%d descriptors successfully retrieved\n", DescriptorCount);

		do
		{
			ShowDescriptorContent(Descriptors[DescriptorCount - 1].Selector, Descriptors[DescriptorCount - 1].Descriptor);

			// Make a change to the descriptor
			Descriptors[DescriptorCount - 1].Descriptor.SegmentType.DataSegment.ExpandDown = !Descriptors[DescriptorCount - 1].Descriptor.SegmentType.DataSegment.ExpandDown;
		} while (DescriptorCount--);

		DescriptorCount = 2;

		if (!SetMultipleDescriptors(&DescriptorCount, &(Descriptors[0].Selector)))
		{
			printf("%d descriptors successfully set\n", DescriptorCount);

			DescriptorCount = 2;
			
			if (!GetMultipleDescriptors(&DescriptorCount, &(Descriptors[0].Selector)))
			{
				printf("%d descriptors successfully retrieved\n", DescriptorCount);

				do
				{
					ShowDescriptorContent(Descriptors[DescriptorCount - 1].Selector, Descriptors[DescriptorCount - 1].Descriptor);
				} while (DescriptorCount--);
			}
			else
			{
				printf("Only re-retrieved %d descriptors\n", DescriptorCount);
			}
		}
		else
		{
			printf("Only set %d descriptor\n", DescriptorCount);
		}
	}
	else
	{
		printf("Only managed to get %d descriptor\n", DescriptorCount);
	}

	if (!FreeLDTDescriptor(Descriptors[0].Selector))
	{
		printf("Alias descriptor freed\n");
	}
	else
	{
		printf("Failed to free alias descriptor\n");
	}
	if (!FreeLDTDescriptor(Descriptors[1].Selector))
	{
		printf("Alias descriptor freed\n");
	}
	else
	{
		printf("Failed to free alias descriptor\n");
	}
}


void TestDOSAllocation(void)
{
	WORD wLargestBlock;
	SEGMENT ASegment;
	SELECTOR ASelector;
	
	if (!AllocateDOSMemoryBlock(0x10, &ASegment, &ASelector, &wLargestBlock))
	{
		printf("DOS memory block allocated\n");
		printf(" Segment : %04X\n", ASegment);
		printf("Selector : %04X\n", ASelector);
		ShowDescriptorDetail(ASelector);

		if (!ResizeDOSMemoryBlock(ASelector, 0x1010, &wLargestBlock))
		{
			printf("DOS memory block resized\n");
			ShowDescriptorDetail(ASelector);
		}
		else
		{
			printf("Failed to resize DOS memory block\n");
			printf("Largest Free Block : %D paragraphs\n", wLargestBlock);
		}

		if (FreeDOSMemoryBlock(ASelector))
		{
			printf("Failed to free DOS memory block");
		}
	}
	else
	{
		printf("Failed to allocate DOS memory block\n");
		printf("Largest Free Block : %D paragraphs\n", wLargestBlock);
	}
}


void main(void)
{
	WORD wSelIncVal;

	Copyright();
	
	if (TestCPUMode())
	{
		wSelIncVal = GetSelectorIncrementValue();
		printf("Selector increment value is %d\n", wSelIncVal);

		TestLDTSelectors();

		TestCodeSegment();

		TestAliasDescriptor();

		TestRealModeDescriptor();

		TestSpecificLDTDescriptor();

		TestGetAndSetDescriptor();

		TestMultipleGetAndSet();

		TestDOSAllocation();
	}
}
